@@ -10,9 +10,9 @@
 //! This module provides means to grant the parts access to the relevant subsets of such global
 //! data structure while masking the fact it is part of the bigger whole from the component.
 //!
-//! Note that the [`cache`][::cache] module has its own [`Access`][::cache::Access] trait that
-//! serves a similar purpose, but with cached access. The signatures are different, therefore an
-//! incompatible trait.
+//! Note that the [`cache`][crate::cache] module has its own [`Access`][crate::cache::Access] trait
+//! that serves a similar purpose, but with cached access. The signatures are different, therefore
+//! an incompatible trait.
 //!
 //! # The general idea
 //!
@@ -21,15 +21,14 @@
 //! relevant part, without knowing what the big structure is.
 //!
 //! For technical reasons, the [`Access`] trait is not object safe. If type erasure is desired, it
-//! is possible use the [`DynAccess`][::access::DynAccess] instead, which is object safe, but
-//! slightly slower.
+//! is possible use the [`DynAccess`] instead, which is object safe, but slightly slower.
 //!
 //! For some cases, it is possible to use [`ArcSwapAny::map`]. If that is not flexible enough, the
 //! [`Map`] type can be created directly.
 //!
 //! Note that the [`Access`] trait is also implemented for [`ArcSwapAny`] itself. Additionally,
-//! there's the [`Constant`][::access::Constant] helper type, which is useful mostly for testing
-//! (it doesn't allow reloading).
+//! there's the [`Constant`] helper type, which is useful mostly for testing (it doesn't allow
+//! reloading).
 //!
 //! # Performance
 //!
@@ -43,32 +42,32 @@
 //! # Examples
 //!
 //! ```rust
-//! extern crate arc_swap;
-//!
 //! use std::sync::Arc;
-//! use std::thread;
+//! use std::thread::{self, JoinHandle};
 //! use std::time::Duration;
 //!
 //! use arc_swap::ArcSwap;
 //! use arc_swap::access::{Access, Constant, Map};
 //!
-//! fn work_with_usize<A: Access<usize> + Send + 'static>(a: A) {
+//! fn work_with_usize<A: Access<usize> + Send + 'static>(a: A) -> JoinHandle<()> {
 //!     thread::spawn(move || {
-//!         loop {
-//!             let value = a.load();
-//!             println!("{}", *value);
+//!         let mut value = 0;
+//!         while value != 42 {
+//!             let guard = a.load();
+//!             value = *guard;
+//!             println!("{}", value);
 //!             // Not strictly necessary, but dropping the guard can free some resources, like
 //!             // slots for tracking what values are still in use. We do it before the sleeping,
 //!             // not at the end of the scope.
-//!             drop(value);
+//!             drop(guard);
 //!             thread::sleep(Duration::from_millis(50));
 //!         }
-//!     });
+//!     })
 //! }
 //!
 //! // Passing the whole thing directly
 //! // (If we kept another Arc to it, we could change the value behind the scenes)
-//! work_with_usize(Arc::new(ArcSwap::from_pointee(42)));
+//! work_with_usize(Arc::new(ArcSwap::from_pointee(42))).join().unwrap();
 //!
 //! // Passing a subset of a structure
 //! struct Cfg {
@@ -76,20 +75,20 @@
 //! }
 //!
 //! let cfg = Arc::new(ArcSwap::from_pointee(Cfg { value: 0 }));
-//! work_with_usize(Map::new(Arc::clone(&cfg), |cfg: &Cfg| &cfg.value));
+//! let thread = work_with_usize(Map::new(Arc::clone(&cfg), |cfg: &Cfg| &cfg.value));
 //! cfg.store(Arc::new(Cfg { value: 42 }));
+//! thread.join().unwrap();
 //!
 //! // Passing a constant that can't change. Useful mostly for testing purposes.
-//! work_with_usize(Constant(42));
+//! work_with_usize(Constant(42)).join().unwrap();
 //! ```
+use core::marker::PhantomData;
+use core::ops::Deref;
 
-use std::marker::PhantomData;
-use std::ops::Deref;
-use std::rc::Rc;
-use std::sync::Arc;
+use crate::imports::{Arc, Box, Rc};
 
-use super::gen_lock::LockStorage;
 use super::ref_cnt::RefCnt;
+use super::strategy::Strategy;
 use super::{ArcSwapAny, Guard};
 
 /// Abstracts over ways code can get access to a value of type `T`.
@@ -111,59 +110,79 @@ pub trait Access<T> {
     fn load(&self) -> Self::Guard;
 }
 
-impl<T, A: Access<T>, P: Deref<Target = A>> Access<T> for P {
+impl<T, A: Access<T> + ?Sized, P: Deref<Target = A>> Access<T> for P {
     type Guard = A::Guard;
     fn load(&self) -> Self::Guard {
         self.deref().load()
     }
 }
 
-impl<T: RefCnt, S: LockStorage> Access<T> for ArcSwapAny<T, S> {
-    type Guard = Guard<'static, T>;
+impl<T> Access<T> for dyn DynAccess<T> + '_ {
+    type Guard = DynGuard<T>;
+
+    fn load(&self) -> Self::Guard {
+        self.load()
+    }
+}
+
+impl<T> Access<T> for dyn DynAccess<T> + '_ + Send {
+    type Guard = DynGuard<T>;
+
+    fn load(&self) -> Self::Guard {
+        self.load()
+    }
+}
+
+impl<T> Access<T> for dyn DynAccess<T> + '_ + Sync + Send {
+    type Guard = DynGuard<T>;
+
+    fn load(&self) -> Self::Guard {
+        self.load()
+    }
+}
+
+impl<T: RefCnt, S: Strategy<T>> Access<T> for ArcSwapAny<T, S> {
+    type Guard = Guard<T, S>;
 
     fn load(&self) -> Self::Guard {
         self.load()
     }
 }
 
-/// Plumbing type.
-///
-/// Accessible, but not expected to be used directly in general.
 #[derive(Debug)]
-pub struct DirectDeref<T: RefCnt>(Guard<'static, T>);
+#[doc(hidden)]
+pub struct DirectDeref<T: RefCnt, S: Strategy<T>>(Guard<T, S>);
 
-impl<T> Deref for DirectDeref<Arc<T>> {
+impl<T, S: Strategy<Arc<T>>> Deref for DirectDeref<Arc<T>, S> {
     type Target = T;
     fn deref(&self) -> &T {
         self.0.deref().deref()
     }
 }
 
-impl<T, S: LockStorage> Access<T> for ArcSwapAny<Arc<T>, S> {
-    type Guard = DirectDeref<Arc<T>>;
+impl<T, S: Strategy<Arc<T>>> Access<T> for ArcSwapAny<Arc<T>, S> {
+    type Guard = DirectDeref<Arc<T>, S>;
     fn load(&self) -> Self::Guard {
         DirectDeref(self.load())
     }
 }
 
-impl<T> Deref for DirectDeref<Rc<T>> {
+impl<T, S: Strategy<Rc<T>>> Deref for DirectDeref<Rc<T>, S> {
     type Target = T;
     fn deref(&self) -> &T {
         self.0.deref().deref()
     }
 }
 
-impl<T, S: LockStorage> Access<T> for ArcSwapAny<Rc<T>, S> {
-    type Guard = DirectDeref<Rc<T>>;
+impl<T, S: Strategy<Rc<T>>> Access<T> for ArcSwapAny<Rc<T>, S> {
+    type Guard = DirectDeref<Rc<T>, S>;
     fn load(&self) -> Self::Guard {
         DirectDeref(self.load())
     }
 }
 
-/// Plumbing type.
-///
-/// This is the guard of [`DynAccess`] trait. It is effectively `Box<Deref<Target = T>>`.
-pub struct DynGuard<T: ?Sized>(Box<Deref<Target = T>>);
+#[doc(hidden)]
+pub struct DynGuard<T: ?Sized>(Box<dyn Deref<Target = T>>);
 
 impl<T: ?Sized> Deref for DynGuard<T> {
     type Target = T;
@@ -183,17 +202,11 @@ impl<T: ?Sized> Deref for DynGuard<T> {
 /// # Examples
 ///
 /// ```rust
-/// extern crate arc_swap;
-///
-/// use std::thread;
-///
 /// use arc_swap::access::{Constant, DynAccess};
 ///
 /// fn do_something(value: Box<dyn DynAccess<usize> + Send>) {
-///     thread::spawn(move || {
-///         let v = value.load();
-///         println!("{}", *v);
-///     });
+///     let v = value.load();
+///     println!("{}", *v);
 /// }
 ///
 /// do_something(Box::new(Constant(42)));
@@ -213,10 +226,70 @@ where
     }
 }
 
-/// A plumbing type.
+/// [DynAccess] to [Access] wrapper.
 ///
-/// This is the guard type for [`Map`]. It is accessible and nameable, but is not expected to be
-/// generally used directly.
+/// In previous versions, `Box<dyn DynAccess>` didn't implement [Access], to use inside [Map] one
+/// could use this wrapper. Since then, a way was found to solve it. In most cases, this wrapper is
+/// no longer necessary.
+///
+/// This is left in place for two reasons:
+/// * Backwards compatibility.
+/// * Corner-cases not covered by the found solution. For example, trait inheritance in the form of
+///   `Box<dyn SomeTrait>` where `SomeTrait: Access` doesn't work out of the box and still needs
+///   this wrapper.
+///
+/// # Examples
+///
+/// The example is for the simple case (which is no longer needed, but may help as an inspiration).
+///
+/// ```rust
+/// use std::sync::Arc;
+///
+/// use arc_swap::ArcSwap;
+/// use arc_swap::access::{AccessConvert, DynAccess, Map};
+///
+/// struct Inner {
+///     val: usize,
+/// }
+///
+/// struct Middle {
+///     inner: Inner,
+/// }
+///
+/// struct Outer {
+///     middle: Middle,
+/// }
+///
+/// let outer = Arc::new(ArcSwap::from_pointee(Outer {
+///     middle: Middle {
+///         inner: Inner {
+///             val: 42,
+///         }
+///     }
+/// }));
+///
+/// let middle: Arc<dyn DynAccess<Middle>> =
+///     Arc::new(Map::new(outer, |outer: &Outer| &outer.middle));
+/// let inner: Arc<dyn DynAccess<Inner>> =
+///     Arc::new(Map::new(AccessConvert(middle), |middle: &Middle| &middle.inner));
+/// let guard = inner.load();
+/// assert_eq!(42, guard.val);
+/// ```
+pub struct AccessConvert<D>(pub D);
+
+impl<T, D> Access<T> for AccessConvert<D>
+where
+    D: Deref,
+    D::Target: DynAccess<T>,
+{
+    type Guard = DynGuard<T>;
+
+    fn load(&self) -> Self::Guard {
+        self.0.load()
+    }
+}
+
+#[doc(hidden)]
 #[derive(Copy, Clone, Debug)]
 pub struct MapGuard<G, F, T, R> {
     guard: G,
@@ -286,10 +359,7 @@ where
     }
 }
 
-/// A plumbing type.
-///
-/// This is the guard type for [`Constant`]. It is accessible, but is not expected to be generally
-/// used directly.
+#[doc(hidden)]
 #[derive(Copy, Clone, Debug, Eq, PartialEq, Ord, PartialOrd, Hash)]
 pub struct ConstantDeref<T>(T);
 
@@ -341,11 +411,11 @@ mod tests {
         check_static_dispatch(a);
     }
 
-    fn check_dyn_dispatch_direct(a: &DynAccess<usize>) {
+    fn check_dyn_dispatch_direct(a: &dyn DynAccess<usize>) {
         assert_eq!(42, *a.load());
     }
 
-    fn check_dyn_dispatch(a: &DynAccess<Arc<usize>>) {
+    fn check_dyn_dispatch(a: &dyn DynAccess<Arc<usize>>) {
         assert_eq!(42, **a.load());
     }
 
@@ -412,7 +482,7 @@ mod tests {
     #[test]
     fn constant() {
         let c = Constant(42);
-        check_static_dispatch_direct(&c);
+        check_static_dispatch_direct(c);
         check_dyn_dispatch_direct(&c);
         check_static_dispatch_direct(c);
     }
@@ -425,4 +495,49 @@ mod tests {
         a.store(Arc::new(Cfg { value: 42 }));
         assert_eq!(42, *Access::load(&map));
     }
+
+    // Compile tests for dynamic access
+    fn _expect_access<T>(_: impl Access<T>) {}
+
+    fn _dyn_access<T>(x: Box<dyn DynAccess<T> + '_>) {
+        _expect_access(x)
+    }
+
+    fn _dyn_access_send<T>(x: Box<dyn DynAccess<T> + '_ + Send>) {
+        _expect_access(x)
+    }
+
+    fn _dyn_access_send_sync<T>(x: Box<dyn DynAccess<T> + '_ + Send + Sync>) {
+        _expect_access(x)
+    }
+
+    #[test]
+    #[allow(clippy::arc_with_non_send_sync)] // Whatever, it's tests...
+    fn double_dyn_access_complex() {
+        struct Inner {
+            val: usize,
+        }
+
+        struct Middle {
+            inner: Inner,
+        }
+
+        struct Outer {
+            middle: Middle,
+        }
+
+        let outer = Arc::new(ArcSwap::from_pointee(Outer {
+            middle: Middle {
+                inner: Inner { val: 42 },
+            },
+        }));
+
+        let middle: Arc<dyn DynAccess<Middle>> =
+            Arc::new(Map::new(outer, |outer: &Outer| &outer.middle));
+        let inner: Arc<dyn DynAccess<Inner>> =
+            Arc::new(Map::new(middle, |middle: &Middle| &middle.inner));
+        // Damn. We have the DynAccess wrapper in scope and need to disambiguate the inner.load()
+        let guard = Access::load(&inner);
+        assert_eq!(42, guard.val);
+    }
 }
